import numpy as np
from PIL import Image, ImageDraw, ImageChops

from tspart._helpers import line_angle, circle_point


def draw_multi_thickness_line(draw, xy, widths, fill=None):
    if len(xy) != 2:
        raise ValueErrror("xy must be of length 2")
    if len(widths) != 2:
        raise ValueErrror("widths must be of length 2")

    angle = line_angle(xy[0], xy[1])

    point_transforms = []
    for point, width in zip(xy, widths):
        radius = width / 2
        point_transforms.append(circle_point(radius, angle + (np.pi / 2), point))
        point_transforms.append(circle_point(radius, angle - (np.pi / 2), point))

    polygon_points = point_transforms[:2] + point_transforms[2:][::-1]
    polygon_points = [tuple(_) for _ in polygon_points]

    draw.polygon(
        xy=polygon_points,
        fill=fill,
        outline=None
    )


def draw_route(
        points,
        factors,
        size,
        line_width=2,
        minimum_line_width_factor=0.05,
        scale=1,
        closed=True,
        background=(255, 255, 255),
        foreground=(0, 0, 0),
        subpixels=8
):
    size = np.array(size)

    subpixels = max(1, subpixels)

    size_scale = tuple((size * subpixels * scale).round().astype(int))

    size_out = tuple((size * scale).round().astype(int))

    line_width = line_width * subpixels * scale

    img = Image.new(mode="RGB", size=size_scale, color=background)
    draw = ImageDraw.Draw(img)

    last_point = None
    last_width = None

    def draw_line_to_point(p, f=None, cap=True):
        nonlocal last_point, last_width

        p = tuple((p * subpixels * scale))

        f_p = ((1 - minimum_line_width_factor) * f) + minimum_line_width_factor
        w = f_p * line_width

        r = w / 2

        # Draw dot
        if cap:
            draw.ellipse(
                xy=(
                    (round(p[0] - r), round(p[1] - r)),
                    (round(p[0] + r), round(p[1] + r))
                ),
                fill=foreground,
                outline=None
            )

        # Draw line
        if last_point is not None:
            draw_multi_thickness_line(
                draw=draw,
                xy=(last_point, p),
                widths=(last_width, w),
                fill=foreground
            )

        last_point = p
        last_width = w

    for point, factor in zip(points, factors):
        draw_line_to_point(point, factor)

    if closed:
        draw_line_to_point(points[0], factors[0], cap=False)

    img = img.resize(size_out, resample=Image.Resampling.LANCZOS)

    return img


def draw_cmyk_routes(
        cmyk_points,
        cmyk_factors,
        size=None,
        line_width=2,
        minimum_line_width_factor=0.05,
        scale=1,
        closed=True,
        subpixels=8
):
    size = np.array(size)

    size_out = tuple((size * scale).round().astype(int))

    sub_colors = (
        (255, 0, 0),
        (0, 255, 0),
        (0, 0, 255),
        (255, 255, 255)
    )

    img = Image.new(size=size_out, mode="RGB", color=(255, 255, 255))
    for points, factor, color in zip(cmyk_points, cmyk_factors, sub_colors):
        channel_img = draw_route(
            points=points,
            factors=factor,
            size=size,
            line_width=line_width,
            minimum_line_width_factor=minimum_line_width_factor,
            scale=scale,
            closed=closed,
            background=(0, 0, 0),
            foreground=color,
            subpixels=subpixels
        ).convert("RGB")

        img = ImageChops.subtract(img, channel_img)

    return img


def draw_rgb_routes(
        rgb_points,
        rgb_factors,
        size=None,
        line_width=2,
        minimum_line_width_factor=0.05,
        scale=1,
        closed=True,
        subpixels=8
):
    size = np.array(size)

    size_out = tuple((size * scale).round().astype(int))

    # TODO: Why are the channels mixed up?
    add_colors = (
        (0, 0, 255),
        (255, 0, 0),
        (0, 255, 0),
    )

    img = Image.new(size=size_out, mode="RGB", color=(0, 0, 0))
    for points, factor, color in zip(rgb_points, rgb_factors, add_colors):
        channel_img = draw_route(
            points=points,
            factors=factor,
            size=size,
            line_width=line_width,
            minimum_line_width_factor=minimum_line_width_factor,
            scale=scale,
            closed=closed,
            background=(0, 0, 0),
            foreground=color,
            subpixels=subpixels
        ).convert("RGB")

        img = ImageChops.add(img, channel_img)

    return img



"""TSP Art Module for Python"""

__version__ = "0.4.14"


import tspart.voronoi
import tspart.tsp
import tspart.neos

from tspart._draw import draw_route, draw_cmyk_routes, draw_rgb_routes
from tspart._image import split_rgb, split_cmyk, rgb_array_to_grayscale
from tspart._helpers import (
    image_to_array, map_points_to_tour, map_points_to_tour_multi, image_array_size, factors_from_image,
    factors_from_image_multi
)
from tspart._files import (
    save_tspart, load_tspart, load_image_as_array, save_tsplib, decode_tsplib, load_tsplib, save_cyc_tour,
    load_cyc_tour, save_jobs, load_jobs
)



import numpy as np
import scipy.spatial
import scipy.ndimage


def get_bounding_corners(points):
    points = np.array(points)

    start = np.floor(points.min(0)).astype(int)
    end = np.ceil(points.max(0)).astype(int)

    return start, end


def image_to_array(image, mode):
    return np.asarray(image.convert(mode))


def map_points_to_tour(points, tour):
    return np.array([points[idx] for idx in tour])


def map_points_to_tour_multi(points_list, tours_list):
    result = []
    for idx, points in enumerate(points_list):
        routes = tours_list[idx]

        r = map_points_to_tour(
            points,
            routes
        )

        result.append(r)

    return result


def ndarray_to_array_2d(array):
    return [list([list(__) for __ in _]) for _ in array]


def array_to_ndarray_2d(array):
    return [np.array(_) for _ in array]


def nearest_point_index(points, location):
    distance, index = scipy.spatial.KDTree(points).query(location)

    return int(index)


def image_array_size(img):
    return tuple([int(round(_)) for _ in img.shape[:2]][::-1])


def line_angle(start, end):
    d = np.array(end) - np.array(start)

    return np.arctan2(d[1], d[0])


def circle_point(radius, angle, offset=(0, 0)):
    offset = np.array(offset)

    x = radius * np.cos(angle)
    y = radius * np.sin(angle)

    point = np.array([x, y])

    return point + offset


def factors_from_image(grayscale_array, points, blur_sigma=3):
    height, width = image_array_size(grayscale_array)

    if blur_sigma > 0:
        grayscale_array = scipy.ndimage.gaussian_filter(grayscale_array, sigma=blur_sigma)

    factors = []
    for point in points:
        x, y = np.array(point).round().astype(int)

        x = min(width - 1, x)
        y = min(height - 1, y)

        factors.append(grayscale_array[x][y] / 255)

    return np.array(factors)


def factors_from_image_multi(grayscale_arrays, points_list, blur_sigma=3):
    factors_list = []
    for grayscale_array, points in zip(grayscale_arrays, points_list):
        factors = factors_from_image(
            grayscale_array=grayscale_array,
            points=points,
            blur_sigma=blur_sigma
        )

        factors_list.append(factors)

    return factors_list



import json
import numpy as np
from PIL import Image

from tspart._helpers import ndarray_to_array_2d, image_to_array, array_to_ndarray_2d


class TsplibSyntaxError(SyntaxError):
    pass


def save_json(filename, obj, indent=2):
    with open(filename, "w") as f:
        json.dump(obj, f, indent=indent)


def load_json(filename):
    with open(filename, "r") as f:
        obj = json.load(f)

    return obj


def save_tspart(filename, points, factors, size, indent=2):
    points = ndarray_to_array_2d(points)
    factors = [list(_) for _ in factors]

    obj = {"points": points, "factors": factors, "size": size}

    save_json(filename=filename, obj=obj, indent=indent)


def load_tspart(filename):
    obj = load_json(filename)

    points = array_to_ndarray_2d(obj["points"])
    factors = [np.array(_) for _ in obj["factors"]]
    size = obj["size"]

    return {"points": points, "factors": factors, "size": size}


def load_image_as_array(filename, mode="RGB"):
    img = Image.open(filename)
    return image_to_array(img, mode)


def make_tsplib(points):
    result = "NAME: tspart\r\n"
    result += "TYPE: TSP\r\n"
    result += f"DIMENSION: {len(points)}\r\n"
    result += "EDGE_WEIGHT_TYPE: EUC_2D\r\n"
    result += "NODE_COORD_SECTION\r\n"

    for idx, (x, y) in enumerate(points):
        result +=f"{idx + 1} {x:.6f} {y:.6f}\r\n"

    return result


def save_tsplib(filename, points):
    with open(filename, "w") as f:
        f.write(make_tsplib(points))


def decode_tsplib(text):
    lines = text.split("\n")
    lines = [_.strip().strip("\r").strip() for _ in lines]

    lines = [_ for _ in lines if _ != ""]

    start_string = "EDGE_WEIGHT_TYPE: EUC_2D"
    if start_string not in lines:
        raise TsplibSyntaxError(f"Invalid format, string '{start_string}' not found")

    start_idx = lines.index(start_string) + 1

    points = []
    for line in lines[start_idx:]:
        split_line = line.split(" ")
        if not all([_.isnumeric() for _ in split_line]):
            break

        points.append(np.array([split_line[1], split_line[2]]))

    return points


def load_tsplib(filename):
    with open(filename, "r") as f:
        text = f.read()

    return decode_tsplib(text)


def save_cyc_tour(filename, tour):
    with open(filename, "w") as f:
        for idx in tour:
            f.write(f"{idx}\n")


def load_cyc_tour(filename):
    with open(filename, "r") as f:
        tour = [int(_.strip()) for _ in f.readlines() if _.strip() != ""]

    return tour


def save_jobs(filename, jobs):
    save_json(filename, {"jobs": jobs})


def load_jobs(filename):
    return load_json(filename)["jobs"]



import numpy as np


def split_cmyk(rgb_array, threshhold=1):
    data = rgb_array.astype(float) / 255
    threshold = threshhold / 255

    channel_max = data.max(2)
    channel_max[channel_max < threshold] = threshold

    k = 1 - channel_max
    c = (1 - data[:, :, 0] - k) / channel_max
    m = (1 - data[:, :, 1] - k) / channel_max
    y = (1 - data[:, :, 2] - k) / channel_max

    result = 1 - np.array([c, m, y, k])

    return tuple([(_ * 255).round().astype(int) for _ in result])


def split_rgb(rgb_array):
    r = rgb_array[:, :, 0]
    g = rgb_array[:, :, 1]
    b = rgb_array[:, :, 2]

    return r, g, b


def rgb_to_gray(rgb):
    r, g, b = rgb

    return int(round((0.299 * r) + (0.587 * g) + (0.114 * b)))


def rgb_array_to_grayscale(rgb_array):
    return np.array([[rgb_to_gray(__) for __ in _] for _ in rgb_array])
